/*****************************************************************************/
/*  Copyright (c) 2008 NXP B.V.  All rights are reserved.                    */
/*  Reproduction in whole or in part is prohibited without the prior         */
/*  written consent of the copyright owner.                                  */
/*                                                                           */
/*  This software and any compilation or derivative thereof is, and          */
/*  shall remain the proprietary information of NXP and is                   */
/*  highly confidential in nature. Any and all use hereof is restricted      */
/*  and is subject to the terms and conditions set forth in the              */
/*  software license agreement concluded with NXP B.V.                       */
/*                                                                           */
/*  Under no circumstances is this software or any derivative thereof        */
/*  to be combined with any Open Source Software, exposed to, or in any      */
/*  way licensed under any Open License Terms without the express prior      */
/*  written permission of the copyright owner.                               */
/*                                                                           */
/*  For the purpose of the above, the term Open Source Software means        */
/*  any software that is licensed under Open License Terms. Open             */
/*  License Terms means terms in any license that require as a               */
/*  condition of use, modification and/or distribution of a work             */
/*                                                                           */
/*  1. the making available of source code or other materials                */
/*     preferred for modification, or                                        */
/*                                                                           */
/*  2. the granting of permission for creating derivative                    */
/*     works, or                                                             */
/*                                                                           */
/*  3. the reproduction of certain notices or license terms                  */
/*     in derivative works or accompanying documentation, or                 */
/*                                                                           */
/*  4. the granting of a royalty-free license to any party                   */
/*     under Intellectual Property Rights                                    */
/*                                                                           */
/*  regarding the work and/or any work that contains, is combined with,      */
/*  requires or otherwise is based on the work.                              */
/*                                                                           */
/*  This software is provided for ease of recompilation only.                */
/*  Modification and reverse engineering of this software are strictly       */
/*  prohibited.                                                              */
/*                                                                           */
/*****************************************************************************/
/**
    @file   ak470x.c
    @brief  Linux implementation of the AK470X (AV Scart switch) driver.

    @b  Module:  AK470X (AV Scart switch) driver.

*//*

Rev Date        Author      Comments
--------------------------------------------------------------------------------
  1 20060808	griffiths   Removed backend class, modified defaults for STB225 & IFA demo
  3 20061115    neill       Changes for audio implementation
  4 20061127    neill       Create as a miscellaneous device
  5 20070112    steel       update for linux_VARIANT support
  6 20071002    neill       Add support for YC output
  7 20080205    scully      Fix popping on boot.
6.1.1 20080206  batelaan    Update file copyright/license text
  8 20080207    scully      Merge.
  9 20080930    sjones      Add power management
 10 20080930    sjones      Replace magic numbers with macros. Save status register state on suspend.
--------------------------------------------------------------------------------
    For consistency and standardisation retain the Section Separators.
*/

#define __AK470X_C__

#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/kmod.h>
#include <linux/miscdevice.h>

#include <asm/io.h>
#include <asm/uaccess.h>

#include <linux/i2c.h>
#include <linux/i2c-dev.h>

//  ----------------------- AK470X MODULE Local Headers --------------------
#include "ak470x.h"
#include "i2c-id.h"
//BPG removed analog backend class
// #include "analog_backend.h"

//  ----------------------- AK470X MODULE Generic Decriptions   ----------------
#define THIS_MODULE_DESCRIPTION     AK470X_MODULE_DESCRIPTION
MODULE_DESCRIPTION(THIS_MODULE_DESCRIPTION);
MODULE_AUTHOR("Mike Neill - Alain Clement");
MODULE_LICENSE("Proprietary");

//  ----------------------- AK470X MODULE Debug Level   ------------------------
static int debug = 1;
module_param(debug, int, 0);
MODULE_PARM_DESC(debug, "Debug level (0-1)");

//  ----------------------- AK470X MODULE Loopthrough   ------------------------
static int loopthrough = 0;
module_param(loopthrough, int, 0);
MODULE_PARM_DESC(loopthrough, "Loopthrough off(0) or On (1)");


//  ----------------------- AK470X Kernel Debug --------------------------------
#ifdef __AK470X_C__
#undef  dprintk
#define dprintk(num, format, args...)                                           \
    do {                                                                        \
        if (debug >= num) printk(format, ##args);                               \
    } while (0)
#endif

//------------------------------------------------------------------------------
//                              FORWARD FUNCTIONS DECLARATION
//------------------------------------------------------------------------------
//  ----------------------- AK470X MODULE Forward Functions Declarations    ----
static int ak470x_attach_adapter    (struct i2c_adapter *adapter    );
static int ak470x_detach_client     (struct i2c_client *client      );
static int ak470x_command           (struct inode  *inode, struct file   *file,
                                     unsigned int   cmd, unsigned long  arg );
#ifdef CONFIG_PM
static int ak470x_suspend(struct i2c_client *client, pm_message_t mesg);
static int ak470x_resume(struct i2c_client *client);
#endif
//  ****************************************************************************
//  *                                                                          *
//  *                           M O D U L E    S T R U C T U R E S             *
//  *                                                                          *
//  ****************************************************************************
//  ----------------------- AK470X MODULE Generic Structure ----------------
static struct file_operations ak470x_fops = {
     .ioctl =  ak470x_command,
};

static struct miscdevice ak470x_device = {
     .minor =    MISC_DYNAMIC_MINOR,
     .name  =    "ak470x",
     .fops  =   &ak470x_fops
};
//  ****************************************************************************
//  *                                                                          *
//  *                           I 2 C    S T R U C T U R E S                   *
//  *                                                                          *
//  ****************************************************************************
#define I2C_AK470X  0x22
#define I2C_NAME(s) (s)->name

//  -----------------------------------------------------------------------------
//                          I2C Support Structures
//  -----------------------------------------------------------------------------
// Generic i2c probe ... addresses: i2c wants 7 bit (without the r/w bit), so '>>1'
static unsigned short   normal_i2c      [ ] =   { I2C_AK470X >> 1,I2C_CLIENT_END};
static unsigned short   probe           [2] =   { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short   ignore          [2] =   { I2C_CLIENT_END, I2C_CLIENT_END };

static struct i2c_client_address_data   addr_data = {
    .normal_i2c         =   normal_i2c,
    .probe              =   probe,
    .ignore             =   ignore,
};

static int    is_ak4702 = 0;
static struct i2c_client *ak470x_client = NULL;
static struct i2c_driver i2c_driver_ak470x = {
    .driver = {
       .name = "ak470x",
    },
    .id = I2C_DRIVERID_AK470X,
    .attach_adapter = ak470x_attach_adapter,
    .detach_client = ak470x_detach_client,
#ifdef CONFIG_PM
      .suspend = ak470x_suspend,
      .resume = ak470x_resume,
#endif
};

//  ****************************************************************************
//  *                                                                          *
//  *                           S W I T C H    S T R U C T U R E S             *
//  *                                                                          *
//  ****************************************************************************
//  ----------------------- AK470X AV Switch Registers Address Mapping  --------
typedef enum    AK470X_SWITCH_REGISTER      {
        AK470X_ADDR_CONTROL             =   0
    ,   AK470X_ADDR_MAIN_SWITCH         =   1
    ,   AK470X_ADDR_VOLUME              =   2
    ,   AK470X_ADDR_ZEROCROSS           =   3
    ,   AK470X_ADDR_VIDEO_SWITCH        =   4
    ,   AK470X_ADDR_VIDEO_OUTPUT        =   5
    ,   AK470X_ADDR_VIDEO_VOLUME        =   6
    ,   AK470X_ADDR_BLANKING_CONTROL    =   7
    ,   AK470X_ADDR_BLANKING_MONITOR    =   8
    ,   AK4706_ADDR_MONITOR_MASK        =   9
    ,   AK4706_ADDR_HD_SWITCH           =   10
    ,   AK4706_ADDR_HD_FILTER           =   11
}   AK470X_SWITCH_REGISTER_t;

//  ----------------------- AK470X AV Switch reg. Indexing  --------------------
typedef enum    AK470X_SWITCH_INDEX_t   {
        AK470X_START_REG        =   0
    ,   AK470X_CONTROL          =   1
    ,   AK470X_MAIN_SWITCH      =   2
    ,   AK470X_VOLUME           =   3
    ,   AK470X_ZEROCROSS        =   4
    ,   AK470X_VIDEO_SWITCH     =   5
    ,   AK470X_VIDEO_OUTPUT     =   6
    ,   AK470X_VIDEO_VOLUME     =   7
    ,   AK470X_BLANKING_CONTROL =   8
    ,   AK470X_BLANKING_MONITOR =   9
    ,   AK4702_NR_REG           =   9
    ,   AK4706_MONITOR_MASK     =   10
    ,   AK4706_HD_SWITCH        =   11
    ,   AK4706_HD_FILTER        =   12
    ,   AK4706_NR_REG           =   12
}   AK470X_SWITCH_INDEX_t;

#ifdef CONFIG_PM
#define AK470X_POWER_DOWN 0x7b
static int keep_control_status = 0;
#endif

//  ----------------------------------------------------------------------------
//                          AK470X AV Switch Local Data
//  ----------------------------------------------------------------------------
//  ----------------------- AK470X AV Switch Data Default: loop through     ----

static av_switch_data av_def_info =
{ 0,
    {
        { SRC_PRIMARY, MODE_YC, BLANK_SLOW_OFF},
        { SRC_PRIMARY, MODE_CVBS, BLANK_SLOW_OFF}
    }
};

//  ----------------------- AK470X AV Switch Back Buffer    --------------------
u8  ak470x_RegBackBuffer[AK4706_NR_REG+1];

//  ----------------------- AK470X AV Switch - Current Settings     ------------
static av_switch_data av_current_info;

/**
 * @brief Is the scart switch installed and supported on this board?
 */
static int ak470x_installed = 1;
static int ak470x_initialised = 0;

//  ****************************************************************************
//  *                                                                          *
//  *                           I 2 C    S E C T I O N                         *
//  *                                                                          *
//  ****************************************************************************
//  -----------------------------------------------------------------------------
//                          I2C Support Functions
//  -----------------------------------------------------------------------------
//  ----------------------- I2C Video Support Function  (Write)     ------------
static int ak470x_write_block ( struct i2c_client* client, const u8* data,
                                unsigned int len)
{
    int ret = -1;
    u8 reg = *data;     /* first register to write to */

    /* Sanity check */
    if (reg + (len - 1) > AK4706_NR_REG)
    {
        return ret;
    }

    if (client == NULL) 
    {
        return 0;
    }

    /* the ak470x has an autoincrement function, use it if
     * the adapter understands raw I2C */
    if (i2c_check_functionality(client->adapter, I2C_FUNC_I2C))
    {
        struct i2c_msg msg;

        msg.len = len;
        msg.buf = (char *) data;
        msg.addr = client->addr;
        msg.flags = 0;
        ret = i2c_transfer(client->adapter, &msg, 1);
    }
    else
    {
        for (++data, --len; len; len--)
        {
            if ((ret = i2c_smbus_write_byte_data(client, reg++, *data++)) < 0)
            {
                break;
            }
        }
    }

    return ret;
}

//  ----------------------- I2C Video Support Function  (Read )     ------------
static inline int ak470x_read (struct i2c_client *client, u8 addr)
{
    if (client == NULL) 
    {
        return 0;
    }

    return i2c_smbus_read_byte_data(client, addr);
}

//  ----------------------- I2C Video Support Function  (Write )    ------------
static inline int ak470x_write (struct i2c_client *client, u8 addr, u8 byte)
{
    u8 data[2];
    if (client == NULL) 
    {
        return 0;
    }

    data[0] = addr;
    data[1] = byte;

    return ak470x_write_block ( client, data,2);
}

//  ----------------------- I2C Audio Support Function  (Read )     ------------
static int ak470x_audio_block_read  ( struct i2c_client *client,
                 u8              addr,          // register address offset
                 u8             *data,          // data buffer ptr    (written to)
                 unsigned int    len)           // data buffer length (bytes)
{
    int ret = -1;

    if (client == NULL) 
    {
        return 0;
    }

    dprintk(1,
        KERN_INFO
        "ak470x.c: Reading %d bytes from address 0x%0X starting at register %d\n",
        len, client->addr, addr);

    while (len--)   {
        *data++ = i2c_smbus_read_byte_data (client, addr++);
    }

    ret = 0;

    return ret;
}

//  ----------------------- I2C Audio Support Function  (Write)     ------------
static int ak470x_audio_block_write ( struct i2c_client *client,
                 u8              addr,          // register address offset
                 const u8       *data,          // data buffer ptr    (read from)
                 unsigned int    len)           // data buffer length (bytes)
{
    int ret = -1;

    if (client == NULL) 
    {
        return 0;
    }

    dprintk(1,
        KERN_INFO
        "ak470x.c: Writing %d bytes to address 0x%0X starting at register %d\n",
        len, client->addr, addr);

    while (len--)   {
        if ( (ret = ak470x_write (client, addr++, *data++) ) < 0) {
            break;
        }
    }

    return ret;
}
//  ****************************************************************************
//  *                                                                          *
//  *                           S W I T C H    S E C T I O N                   *
//  *                                                                          *
//  ****************************************************************************
//  ----------------------------------------------------------------------------
//                          Switch control functions
//  ----------------------------------------------------------------------------
//  ----------------------- AK470X Mode Function (Write/Post BackBuffer) -------
static int ak470x_switch_post_regbackbuffer (struct i2c_client *client)
{
    return ak470x_write_block (client, ak470x_RegBackBuffer, is_ak4702 ? AK4702_NR_REG+1 : AK4706_NR_REG+1);
}

//  ----------------------- AK470X Mode Function (Loopthrough) -----------------
static void ak470x_switch_set_loopthrough (int loopthrough)
{
    // Loopthrough Mode
    if (loopthrough)
    {
    	ak470x_RegBackBuffer[AK470X_CONTROL] |=  0x01;
    }
    else
    {
    	ak470x_RegBackBuffer[AK470X_CONTROL] &= ~0x01;
    }
}

//  ----------------------- AK470X Mode Function (Global Switch Setup)  --------
int ak470x_switch_set (const av_switch_data* pInfo)
{
    // Store settings
    av_current_info = *pInfo;

    // Loopthrough Mode
    ak470x_switch_set_loopthrough (pInfo->loopthrough);
    //
    // Reset all register values to zero
    //
    // - Preserve 'ak470x_RegBackBuffer[AK470X_VOLUME]'
    // - Preserve 'ak470x_RegBackBuffer[AK470X_ZEROCROSS]'
    //
    printk(" Alles Register werden null gesetzt.\n");
    ak470x_RegBackBuffer[AK470X_START_REG]          =   0x00;
    ak470x_RegBackBuffer[AK470X_MAIN_SWITCH]        =   0x00;
    ak470x_RegBackBuffer[AK470X_VIDEO_SWITCH]       =   0x00;
    ak470x_RegBackBuffer[AK470X_VIDEO_OUTPUT]       =   0x00;
    ak470x_RegBackBuffer[AK470X_VIDEO_VOLUME]       =   0x04;
    ak470x_RegBackBuffer[AK470X_BLANKING_CONTROL]   =   0x00;
    ak470x_RegBackBuffer[AK470X_BLANKING_MONITOR]   =   0x00;
    ak470x_RegBackBuffer[AK4706_MONITOR_MASK]       =   0x08;
    ak470x_RegBackBuffer[AK4706_HD_SWITCH]          =   0x00;
    ak470x_RegBackBuffer[AK4706_HD_FILTER]          =   0x2A;

    // TV Output Source Select
    switch(pInfo->attribute[TV_OUTPUT].src_select)
    {
        case(SRC_PRIMARY) :
            switch(pInfo->attribute[TV_OUTPUT].display_mode)
            {
                case(MODE_RGB) :
		    printk("ak470x: MODE RGB eingestellt!");
                    ak470x_RegBackBuffer[AK470X_VIDEO_SWITCH]       |= 0x01;
                    ak470x_RegBackBuffer[AK470X_BLANKING_CONTROL]   |= 0x01;
                    break;
                case(MODE_YC) :
                    ak470x_RegBackBuffer[AK470X_VIDEO_SWITCH]       |= 0x03;
                    break;
                case(MODE_CVBS) :
                    ak470x_RegBackBuffer[AK470X_VIDEO_SWITCH]       |= 0x01;
                    break;
            }
            break;
        case(SRC_SECONDARY) :
            switch(pInfo->attribute[TV_OUTPUT].display_mode)
            {
                case(MODE_RGB) :
                    /* Invalid Mode */
                    dprintk(1, KERN_INFO "AK470X: SECONDARY output RGB mode not possible - Falling back to CVBS\n");
                    /* Fall through! */
                case(MODE_CVBS) :
                    ak470x_RegBackBuffer[AK470X_VIDEO_SWITCH]       |= 0x02;
                    break;
                case(MODE_YC) :
                    ak470x_RegBackBuffer[AK470X_VIDEO_SWITCH]       |= 0x02;
                    break;
            }
            break;
        case(SRC_TVIN) :
            /* Invalid Mode */
            dprintk(1, KERN_INFO "AK470X: TV OUT using TV IN src not possible\n");
            return -EINVAL;
            break;
        case(SRC_VCRIN) :
            /* This mode is used during HD output ... do not route audio from VCR IN! */
            // ak470x_RegBackBuffer[AK470X_MAIN_SWITCH]            |= 0x01;
            ak470x_RegBackBuffer[AK470X_VIDEO_SWITCH]           |= 0x04;
            break;
        case(SRC_MUTE) :
            ak470x_RegBackBuffer[AK470X_MAIN_SWITCH]            |= 0x02;
            break;
    }


    // VCR Output Source Select
    switch(pInfo->attribute[VCR_OUTPUT].src_select)
    {
        case(SRC_PRIMARY) :
            switch(pInfo->attribute[VCR_OUTPUT].display_mode)
            {
                case(MODE_RGB) :
                    /* Invalid Mode */
                    dprintk(1, KERN_INFO "AK470X: VCR RGB mode not possible - Falling back to CVBS\n");
                    /* Fall through! */
                case(MODE_CVBS) :
                    ak470x_RegBackBuffer[AK470X_VIDEO_SWITCH]       |= 0x08;
                    break;
                case(MODE_YC) :
                    ak470x_RegBackBuffer[AK470X_VIDEO_SWITCH]       |= 0x08;
                    break;
            }
            break;
        case(SRC_SECONDARY) :
            /* Mute the audio output - if not using the primary output source */
            // Von mir eingefügt
            ak470x_RegBackBuffer[AK470X_MAIN_SWITCH] |= 0x20;
            switch(pInfo->attribute[VCR_OUTPUT].display_mode)
            {
                case(MODE_RGB) :
                    /* Invalid Mode */
                    dprintk(1, KERN_INFO "AK470X: VCR RGB mode not possible - Falling back to CVBS\n");
                    /* Fall through! */
                case(MODE_CVBS) :
                    ak470x_RegBackBuffer[AK470X_VIDEO_SWITCH]       |= 0x10;
                    break;
                case(MODE_YC) :
                    ak470x_RegBackBuffer[AK470X_VIDEO_SWITCH]       |= 0x10;
                    break;
            }
            break;
        case(SRC_TVIN) :
            ak470x_RegBackBuffer[AK470X_MAIN_SWITCH]            |= 0x10;
            ak470x_RegBackBuffer[AK470X_VIDEO_SWITCH]           |= 0x18;
            break;
        case(SRC_VCRIN) :
            /* Invalid Mode */
            dprintk(1, KERN_INFO "AK470X: VCR OUT using VCR IN src not possible\n");
            return -EINVAL;
            break;
        case(SRC_MUTE) :
            ak470x_RegBackBuffer[AK470X_MAIN_SWITCH]            |= 0x20;
            break;
    }

    // TV Output Display Mode Select
    switch(pInfo->attribute[TV_OUTPUT].display_mode)
    {
        case(MODE_RGB) :
            ak470x_RegBackBuffer[AK470X_VIDEO_OUTPUT]           |= 0x4F;
            break;
        case(MODE_YC) :
            ak470x_RegBackBuffer[AK470X_VIDEO_OUTPUT]           |= 0x03;
            break;
        case(MODE_CVBS) :
            ak470x_RegBackBuffer[AK470X_VIDEO_OUTPUT]           |= 0x01;
            break;
    }


    // VCR Output Display Mode Select
    switch(pInfo->attribute[VCR_OUTPUT].display_mode)
    {
        case(MODE_RGB) :
            /* Invalid Mode */
            dprintk(1, KERN_INFO "AK470X: VCR RGB mode not possible - Falling back to CVBS\n");
            /* Fall through! */
        case(MODE_CVBS) :
            ak470x_RegBackBuffer[AK470X_VIDEO_OUTPUT]           |= 0x10;
            break;
        case(MODE_YC) :
            ak470x_RegBackBuffer[AK470X_VIDEO_OUTPUT]           |= 0x30;
            break;
    }


    // Loopthrough Mode Fallback
    if (pInfo->loopthrough)
    {
        ak470x_RegBackBuffer[AK470X_BLANKING_CONTROL]       |= 0xC0;
    }
    else
    {
        // TV Output Slow Blanking Select
        switch(pInfo->attribute[TV_OUTPUT].slow_blanking)
        {
            case(BLANK_SLOW_OFF) :
                break;
            case(BLANK_SLOW_16_9) :
                ak470x_RegBackBuffer[AK470X_BLANKING_CONTROL]   |= 0x04;
                break;
            case(BLANK_SLOW_4_3) :
                ak470x_RegBackBuffer[AK470X_BLANKING_CONTROL]   |= 0x0C;
                break;
            case(BLANK_SLOW_FOLLOW) :
                ak470x_RegBackBuffer[AK470X_BLANKING_CONTROL]   |= 0xC0;
                break;
            case(BLANK_SLOW_MONITOR) :
                dprintk(1, KERN_INFO "AK470X: MONITOR of TV slow blanking not possible\n");
                return -EINVAL;
                break;
        }

        // VCR Output Slow Blanking Select
        switch(pInfo->attribute[VCR_OUTPUT].slow_blanking)
        {
            case(BLANK_SLOW_OFF) :
                break;
            case(BLANK_SLOW_16_9) :
                ak470x_RegBackBuffer[AK470X_BLANKING_CONTROL]   |= 0x10;
                break;
            case(BLANK_SLOW_4_3) :
                ak470x_RegBackBuffer[AK470X_BLANKING_CONTROL]   |= 0x30;
                break;
            case(BLANK_SLOW_FOLLOW) :
                dprintk(1, KERN_INFO "AK470X: FOLLOW for VCR slow blanking not possible\n");
                return -EINVAL;
                break;
            case(BLANK_SLOW_MONITOR) :
                ak470x_RegBackBuffer[AK470X_BLANKING_CONTROL]   |= 0x80;
                break;
        }
    }

    return ak470x_switch_post_regbackbuffer (ak470x_client);
}

//  ----------------------- AK470X Mode Get Function (Global Switch Read)  ----
int ak470x_switch_get (av_switch_data* pInfo)
{
    *pInfo = av_current_info;
    return 0;
}

//  ----------------------- AK470X function to determine if it is supported on this board or not.  ----
int ak470x_switch_installed ()
{
    return ak470x_installed;
}

//  ----------------------- AK470X Mode Function (Global init)  ----------------
static int ak470x_switch_init (struct i2c_client *client)
{
    int i;
    for(i=0; i<AK4702_NR_REG+1; i++)    ak470x_RegBackBuffer[i] = 0;
    printk("ak470x switch_init");
    // Default volume : 0dB
    ak470x_RegBackBuffer[AK470X_CONTROL]            =   0x70;
    ak470x_RegBackBuffer[AK470X_VOLUME]             =   0x22;
    ak470x_RegBackBuffer[AK470X_ZEROCROSS]          =   0x27;

    // Initialise the AV switch loop through mode
    av_def_info.loopthrough = loopthrough;

    return ak470x_switch_set (&av_def_info);
}

//  ****************************************************************************
//  *                                                                          *
//  *                           A U D I O     S E C T I O N                    *
//  *                                                                          *
//  ****************************************************************************
//
//  ----------------------------------------------------------------------------
//                          Misc Utilities
//  ----------------------------------------------------------------------------

//  ----------------------- Channel Number policy   ----------------------------
static inline int   ak470x_audio_channel_number_isvalid  (int channel)
{
    return ((channel >= AK470X_AUDIO_CHANNEL_NUMBER_MIN) && (channel <= AK470X_AUDIO_CHANNEL_NUMBER_MAX));
}

//  ----------------------- Channel Number to index   --------------------------
static inline int   ak470x_audio_channel_number_to_index (int channel)
{
    return (channel - AK470X_AUDIO_CHANNEL_NUMBER_MIN);
}

//  ----------------------------------------------------------------------------
//                          Volume norm conversion support
//  ----------------------------------------------------------------------------

//  ----------------------- AK470X associative volume buffer -------------------
static AK470X_AUDIO_IOCTL_VOLUME_t	ak470x_delta_volume [AK470X_AUDIO_CHANNEL_COUNT] = 
{
	[0 ... (AK470X_AUDIO_CHANNEL_COUNT-1)]=
	{
		.audio_channel 	= 0
	,	.master_left 	= AK470X_AUDIO_INTERNAL_GAIN_MIN
	,	.master_right	= AK470X_AUDIO_INTERNAL_GAIN_MIN
	,	.external_left	= AK470X_AUDIO_EXTERNAL_VOLUME_MIN
	,	.external_right = AK470X_AUDIO_EXTERNAL_VOLUME_MIN
	}
};

//  ----------------------- AK470X enforce strict boundaries policy ------------
static inline int   ak470x_enforce_volume_boundaries (int val, int min_val, int max_val)
{
    return (val < min_val)? min_val : (val > max_val)? max_val : val;
}

//  ----------------------- AK470X internal gain to volume conversion   --------
static inline int   ak470x_standard_volume (AK470X_AUDIO_MASTER_VOLUME_t Gain)
{
    int gain   =  ak470x_enforce_volume_boundaries ((int)Gain, AK470X_gain_min, AK470X_gain_max);
   	return   AK470X_vol_min + (((gain-AK470X_gain_min) * AK470X_vol_mag) / AK470X_gain_mag);
}

//  ----------------------- AK470X volume to internal gain conversion   --------
static inline AK470X_AUDIO_MASTER_VOLUME_t  ak470x_internal_gain (int Volume)
{
    int volume =   ak470x_enforce_volume_boundaries ((int)Volume, AK470X_vol_min, AK470X_vol_max);
    return  (AK470X_AUDIO_MASTER_VOLUME_t) 
    		 AK470X_gain_min + (((volume-AK470X_vol_min) * AK470X_gain_mag) / AK470X_vol_mag);
}

//  ----------------------------------------------------------------------------
//                          I2C Audio I/O support
//  ----------------------------------------------------------------------------

//  ----------------------- I2C Audio Support Function  (Read )     ------------
static inline int ak470x_audio_read  (struct i2c_client *client, u8 addr, u8 *byte_ptr)
{
    return ak470x_audio_block_read  (client, addr, byte_ptr, 1);
}

//  ----------------------- I2C Audio Support Function  (Write )    ------------
static inline int ak470x_audio_write (struct i2c_client *client, u8 addr, u8 *byte_ptr)
{
    return ak470x_audio_block_write (client, addr, byte_ptr, 1);
}

//  ----------------------------------------------------------------------------
//                          Mode control functions
//  ----------------------------------------------------------------------------

//  ----------------------- AK470X Audio Function (Mode)    --------------------
static int ak470x_set_audio_mode(struct i2c_client *client, AK470X_AUDIO_IOCTL_MODE_t *mode)
{
    int     retValue    =   -EINVAL;

    AK470X_AUDIO_CHANNEL_NUMBER_t       channel     =   mode->audio_channel;

    if (ak470x_audio_channel_number_isvalid(channel))
    {
        retValue    =   -EINVAL;    // NOT IMPLEMENTED YET
        if (retValue) {
            dprintk(1,  KERN_INFO "ak470x.c: Failed to setup mode for audio channel %d (0x%X) (Unimplemented)\n",
            (int) channel, retValue);
            return retValue;
        }
    }
    return retValue;
}

//  ----------------------------------------------------------------------------
//                          Volume control functions
//  ----------------------------------------------------------------------------

//  ----------------------- AK470X Audio Function (Volume get)  ------------
static int ak470x_audio_volume_get (struct i2c_client *client, AK470X_AUDIO_IOCTL_VOLUME_t *volume)
{
    AK470X_AUDIO_CHANNEL_NUMBER_t   channel;
    int     retValue    =   -EINVAL;
    u8      data_byte   =   0;

    // Check channel number
    channel     =   volume->audio_channel;
    if (!ak470x_audio_channel_number_isvalid(channel))  return retValue;

    // I2C buffered I/O
    retValue = ak470x_audio_read  (client, AK470X_ADDR_VOLUME, &data_byte);

    if (retValue) {
        dprintk(1,  KERN_INFO "ak470x.c: Failed to get volume for audio channel %d (0x%X)\n", (int) channel, retValue);
        return retValue;
    }

    // Refresh back buffer
    ak470x_RegBackBuffer[AK470X_VOLUME] = data_byte;

    // GET Audio Master Volume (L/R)
    volume->master_left     =   ak470x_RegBackBuffer[AK470X_VOLUME];
    volume->master_right    =   ak470x_RegBackBuffer[AK470X_VOLUME];

    return retValue;
}

//  ----------------------- AK470X Audio Function (Volume set)  ------------
static int ak470x_audio_volume_set (struct i2c_client *client, AK470X_AUDIO_IOCTL_VOLUME_t *volume)
{
    AK470X_AUDIO_CHANNEL_NUMBER_t   channel;
    int     retValue    =   -EINVAL;
    u8      data_byte   =   0;
    int     gain;

    // Check out channel number
    channel     =   volume->audio_channel;
    if (!ak470x_audio_channel_number_isvalid(channel))  return retValue;

    // SET Audio Master Volume (L+R)/2 -> (Global)
    gain = ((int)volume->master_left + (int)volume->master_right) / 2;

    // Enforce boundaries
    data_byte = (u8) ak470x_enforce_volume_boundaries (gain, AK470X_gain_min, AK470X_gain_max);

    // I2C buffered I/O
    retValue = ak470x_audio_write (client, AK470X_ADDR_VOLUME, &data_byte);
    if (retValue) {
        dprintk(1,  KERN_INFO "ak470x.c: Failed to set volume for audio channel %d (0x%X)\n", (int) channel, retValue);
        return retValue;
    }

    // Refresh back buffer
    ak470x_RegBackBuffer[AK470X_VOLUME] = data_byte;


    return retValue;
}

//  ----------------------- AK470X Audio Function (Global init)  ---------------
static int ak470x_audio_init (struct i2c_client *client)
{
    int     retValue    =   -EINVAL;
    int 	channel;
    
    for (channel=AK470X_AUDIO_CHANNEL_NUMBER_MIN; channel<=AK470X_AUDIO_CHANNEL_NUMBER_MAX; channel++)   
    {
    	AK470X_AUDIO_IOCTL_VOLUME_t	*pvolume, volume;
    	pvolume = &ak470x_delta_volume	[ak470x_audio_channel_number_to_index(channel)];

		// Preset associative buffer
		pvolume->audio_channel 	= 	channel;
		pvolume->master_left	=	(AK470X_AUDIO_MASTER_VOLUME_t)  ak470x_internal_gain (AK470X_AUDIO_EXTERNAL_VOLUME_MID);
		pvolume->master_right	=	(AK470X_AUDIO_MASTER_VOLUME_t)  ak470x_internal_gain (AK470X_AUDIO_EXTERNAL_VOLUME_MID);
		pvolume->external_left	=	AK470X_AUDIO_EXTERNAL_VOLUME_MAX;
		pvolume->external_right	=	AK470X_AUDIO_EXTERNAL_VOLUME_MAX;
    	// Default volume : AK470X_AUDIO_EXTERNAL_VOLUME_MID (-60+((34/2)x2) = -26dB std)
		
		// Set volume 
    	retValue = ak470x_audio_volume_set (client, pvolume);
    	if (retValue)	break;
    	
		// Retrieve actual AK470X chip volume 
    	volume.audio_channel	=	channel;
    	retValue = ak470x_audio_volume_get (client, &volume);
    	if (retValue)	break;
    	
    	// Pick associated buffered external value if match	(LEFT)
    	if (volume.master_left!=pvolume->master_left)
    	{
    		// Else resync with chipset value	(LEFT)
    		pvolume->master_left	=	volume.master_left;
    		pvolume->external_left  = 	ak470x_standard_volume (volume.master_left);
    	}
    	
    	// Pick associated buffered external value if match	(RIGHT)
    	if (volume.master_right!=pvolume->master_right)
    	{
    		// Else resync with fresh chipset value	(RIGHT)
     		pvolume->master_right	=	volume.master_right;
    		pvolume->external_right = 	ak470x_standard_volume (volume.master_right);
    	}
	    	
    }
    return retValue;
}
//  ****************************************************************************
//  *                                                                          *
//  *                           M O D U L E    S E C T I O N                   *
//  *                                                                          *
//  ****************************************************************************
//  ----------------------- AK470X IOCTL switch  (Commands) ----------------
static int ak470x_command ( struct inode  *inode, struct file   *file,
                            unsigned int   cmd, unsigned long  arg)
{
    dprintk(1, KERN_INFO "Got AK470X Command : 0x%X\n", cmd);

    switch (cmd)
    {
        //----------------------------------------------------------------------
        //  AV Switch IOCTLs
        //----------------------------------------------------------------------

        case AK470X_IOCTL_SWITCH_MODE_SET:
            return ak470x_switch_set ((const AK470X_SWITCH_IOCTL_MODE_t*)arg);
            break;

        case AK470X_IOCTL_SWITCH_MODE_GET:
            return ak470x_switch_get ((AK470X_SWITCH_IOCTL_MODE_t*)arg);
            break;

        //----------------------------------------------------------------------
        //  Audio IOCTLs
        //----------------------------------------------------------------------

        case AK470X_IOCTL_AUDIO_MODE_SET:
            return ak470x_set_audio_mode    (ak470x_client, (AK470X_AUDIO_IOCTL_MODE_t*)    arg);
            break;

        case AK470X_IOCTL_AUDIO_VOLUME_GET:
            printk("Aktviere AUDIO_VOUME_GET\n");
            return ak470x_audio_volume_get  (ak470x_client, (AK470X_AUDIO_IOCTL_VOLUME_t*)arg);
            break;

        case AK470X_IOCTL_AUDIO_VOLUME_SET:
            return ak470x_audio_volume_set  (ak470x_client, (AK470X_AUDIO_IOCTL_VOLUME_t*)  arg);
            break;
	case AK470X_IOCTL_REGISTER:
            printk("AK470x: Aktiviere Register:");
	    printk("AK470x: Ergebnis: %i",registersetzen(ak470x_client, (AK470X_REGISTERSET_t*) arg));
            break;
        // Unimplemented IOCTL
        default:
            dprintk(1, KERN_INFO "unknown ak470x??(%d)\n",  cmd);
            return -EINVAL;
    }

    return 0;
}

//  ----------------------- AK470X I2C Detection Function (Detect) ------------
static int ak470x_detect_client (struct i2c_adapter *adapter, int address, int kind)
{
    int rv;
    u8 value;

    dprintk(1,
        KERN_INFO
        "ak470x.c: detecting ak470x client on address 0x%x\n",
        address << 1);

    /* Check if the adapter supports the needed features */
    if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_READ_BYTE | I2C_FUNC_SMBUS_WRITE_BYTE_DATA))
    {
        return 0;
    }

    /* Create the I2C client data structure */
    ak470x_client = kmalloc(sizeof(struct i2c_client), GFP_KERNEL);
    if (ak470x_client == 0)
    {
        return -ENOMEM;
    }
    memset(ak470x_client, 0, sizeof(struct i2c_client));
    ak470x_client->addr = address;
    ak470x_client->adapter = adapter;
    ak470x_client->driver = &i2c_driver_ak470x;
    snprintf(I2C_NAME(ak470x_client), sizeof(I2C_NAME(ak470x_client)) - 1, "ak470x");

    /* Attach the AV switch to the I2C adapter */
    rv = i2c_attach_client(ak470x_client);
    if (rv) {
        dprintk(1,
            KERN_INFO
            "ak470x.c: Failed to attach to client on adapter %s (0x%x) - 0x%02X\n",
            I2C_NAME(adapter), adapter->id, rv);
        kfree(ak470x_client);
        return rv;
    }

    if (ak470x_initialised)
    {
        return 0;
    }

    /* Previous way of determining variant caused a popping noise on  */
    /* the output. So be a bit more explicit in the device selection. */
#if !defined (CONFIG_NXP_STB225)
        is_ak4702 = 0;
#else
        is_ak4702 = 1;
#endif

    ak470x_initialised = 1;

    dprintk(1,
        KERN_INFO
        "ak470x.c: found device (%s) on adapter %s (0x%x) - 0x%02X\n",
        is_ak4702 ? "AK4702" : "AK4706", I2C_NAME(adapter), adapter->id, rv);

    return 0;
}

//  ----------------------- AK470X I2C Detection Function (Attach)     --------
static int ak470x_attach_adapter (struct i2c_adapter *adapter)
{
    return i2c_probe(adapter, &addr_data, &ak470x_detect_client);
}

//  ----------------------- AK470X I2C Detection Function (Detach)     --------
static int ak470x_detach_client (struct i2c_client *client)
{
    int err;

    err = i2c_detach_client(client);
    if (err)
    {
        return err;
    }

    kfree(client);

    return 0;
}

#ifdef CONFIG_PM
static int ak470x_suspend(struct i2c_client *client, pm_message_t mesg)
{
    keep_control_status = ak470x_read(ak470x_client, AK470X_ADDR_CONTROL);
    ak470x_write(ak470x_client, AK470X_ADDR_CONTROL, AK470X_POWER_DOWN);
    return 0;
}

static int ak470x_resume(struct i2c_client *client)
{
    ak470x_write(ak470x_client, AK470X_ADDR_CONTROL, keep_control_status);
    return 0;
}
#endif

//  ----------------------- MODULE INIT     ------------------------------------
static int __init ak470x_init (void)
{
    int result;

    ak470x_initialised = 0;

    /* Register a misc device called "ak470x". */
    result = misc_register( &ak470x_device );
    if (result < 0)
    {
        printk("can't register misc device (minor %d)!\n", ak470x_device.minor );
        return result;
    }

    result = i2c_add_driver(&i2c_driver_ak470x);

    // Initialise the AV switch (loop through mode by default)
    ak470x_switch_init (ak470x_client);

    // Initialise the Audio mixer (50% volume by default)
    ak470x_audio_init  (ak470x_client);

    if (result == 0)
    {
        printk ("%s (%s-%s) - Device %s Loopthrough %s\n",
                THIS_MODULE_DESCRIPTION,
                __DATE__,
                __TIME__,
                is_ak4702 ? "AK4702" : "AK4706",
                loopthrough ? "enabled" : "disabled");
        ak470x_installed = 1;
    }
    else
    {
        printk ("%s (%s-%s) - Scart Switch not available - Dummy Driver only.\n",
                THIS_MODULE_DESCRIPTION,
                __DATE__,
                __TIME__);
        ak470x_installed = 0;
    }
    return 0;
}

//  ----------------------- MODULE EXIT     ------------------------------------
static void __exit ak470x_exit (void)
{
    misc_deregister(&ak470x_device);
    i2c_del_driver(&i2c_driver_ak470x);
    printk("Shutting down %s\n", THIS_MODULE_DESCRIPTION);
}
// Meine Funktionen
void registerzeigen(){
return;
}
//Register setzen
int registersetzen(struct i2c_client *client, AK470X_REGISTERSET_t *werte){
// - Preserve 'ak470x_RegBackBuffer[AK470X_VOLUME]'
    // - Preserve 'ak470x_RegBackBuffer[AK470X_ZEROCROSS]'
    //
    printk(" Aktuelle Regiser sind: Control: %x, Volume: %x, %x Mainswitch: %x\n",ak470x_RegBackBuffer[AK470X_CONTROL],ak470x_RegBackBuffer[AK470X_VOLUME],ak470x_RegBackBuffer[AK470X_VIDEO_VOLUME], ak470x_RegBackBuffer[AK470X_MAIN_SWITCH]);
    printk(" Registersetzen.\n");
 //   ak470x_RegBackBuffer[AK470X_CONTROL]            =   0x70;
 //   ak470x_RegBackBuffer[AK470X_VOLUME]		    =   0x22;
 //   ak470x_RegBackBuffer[AK470X_ZEROCROSS]	    =   0x00;
 //   ak470x_RegBackBuffer[AK470X_START_REG]          =   0x00;
 //   ak470x_RegBackBuffer[AK470X_MAIN_SWITCH]        =   0x00;
 //   ak470x_RegBackBuffer[AK470X_VIDEO_SWITCH]       =   0x00;
 //   ak470x_RegBackBuffer[AK470X_VIDEO_OUTPUT]       =   0x00;
 //   ak470x_RegBackBuffer[AK470X_VIDEO_VOLUME]       =   0x04;
 //   ak470x_RegBackBuffer[AK470X_BLANKING_CONTROL]   =   0x00;
 //   ak470x_RegBackBuffer[AK470X_BLANKING_MONITOR]   =   0x00;
 //   ak470x_RegBackBuffer[AK4706_MONITOR_MASK]       =   0x08;
 //   ak470x_RegBackBuffer[AK4706_HD_SWITCH]          =   0x00;
 //   ak470x_RegBackBuffer[AK4706_HD_FILTER]          =   0x2A;
    
   //  return ak470x_write();
  return ak470x_write(client,werte->adresse,werte->wert);
}

//  ****************************************************************************
//  *                                                                          *
//  *                       P U B L I C    O B J E C T S                       *
//  *                                                                          *
//  ****************************************************************************
//  -----------------------------------------------------------------------------
//                          Exported functions
//  -----------------------------------------------------------------------------
//  ----------------------- AK470X IOCTL wrapper  (Commands)    ----------------
int ak470x_exec (unsigned int cmd, void *p_object)
{
    return ak470x_command (NULL, NULL, cmd, (unsigned long) p_object);
}

//  ----------------------- AK470X exported function    ------------------------
int ak470x_std_volume_get       (int channel, int *p_left, int *p_right)
{
    int     retValue    =   -EINVAL;
    AK470X_AUDIO_IOCTL_VOLUME_t volume;
    AK470X_AUDIO_IOCTL_VOLUME_t *pvolume;
    
    pvolume = &ak470x_delta_volume	[ak470x_audio_channel_number_to_index(channel)];

    volume.audio_channel    =   channel;
    volume.master_left      =   AK470X_AUDIO_INTERNAL_GAIN_MIN;
    volume.master_right     =   AK470X_AUDIO_INTERNAL_GAIN_MIN;

    retValue    =   ak470x_exec (AK470X_IOCTL_AUDIO_VOLUME_GET, &volume);

    if (!retValue) // If ok ...
    {
    	// Pick associated buffered external value if match	(LEFT)
    	if (volume.master_left!=pvolume->master_left)
    	{
    		// Else resync with chipset value	(LEFT)
    		pvolume->external_left  = ak470x_standard_volume (volume.master_left);
    	}
    	
    	// Pick associated buffered external value if match	(RIGHT)
    	if (volume.master_right!=pvolume->master_right)
    	{
    		// Else resync with fresh chipset value	(RIGHT)
    		pvolume->external_right = ak470x_standard_volume (volume.master_right);
    	}
    	
        if (p_left  != NULL) *p_left  = pvolume->external_left;
        if (p_right != NULL) *p_right = pvolume->external_right;
    }

    return  retValue;
}

//  ----------------------- AK470X exported function    ------------------------
int ak470x_std_volume_set       (int channel, int left, int right)
{
    int     retValue    =   -EINVAL;
    AK470X_AUDIO_IOCTL_VOLUME_t *pvolume;
    
    pvolume = &ak470x_delta_volume	[ak470x_audio_channel_number_to_index(channel)];

	// Set channel ID
    pvolume->audio_channel    =   channel;

	// Store requested external scale volumes (values associated with internal ones)
    pvolume->external_left    =   left;
    pvolume->external_right   =   right;
    
	// Set internal scale volumes (values associated with external ones)
    pvolume->master_left      =   (AK470X_AUDIO_MASTER_VOLUME_t)  ak470x_internal_gain (left);
    pvolume->master_right     =   (AK470X_AUDIO_MASTER_VOLUME_t)  ak470x_internal_gain (right);

    retValue    =   ak470x_exec (AK470X_IOCTL_AUDIO_VOLUME_SET, pvolume);

    return  retValue;
}

//  ----------------------- AK470X exported function    ------------------------
int ak470x_std_volume_range_get	(int channel, int *p_min, int *p_max)
{
    int     retValue    =   -EINVAL;

    if (ak470x_audio_channel_number_isvalid(channel)) // If channel# ok ...
    {
        if (p_min != NULL) *p_min = AK470X_vol_min;
        if (p_max != NULL) *p_max = AK470X_vol_max;
        retValue = 0;
    }

    return  retValue;
}
//  ****************************************************************************
//  *                                                                          *
//  *                       E X P O R T E D    S Y M B O L S                   *
//  *                                                                          *
//  ****************************************************************************
EXPORT_SYMBOL(ak470x_exec);
EXPORT_SYMBOL(ak470x_std_volume_get);
EXPORT_SYMBOL(ak470x_std_volume_set);
EXPORT_SYMBOL(ak470x_std_volume_range_get);
EXPORT_SYMBOL(ak470x_switch_set);
EXPORT_SYMBOL(ak470x_switch_get);
EXPORT_SYMBOL(ak470x_switch_installed);

//  ----------------------- MODULE MANAGEMENT   --------------------------------
module_init(ak470x_init);
module_exit(ak470x_exit);
